@@ -53,6 +53,10 @@
 #include "regs.h"
 #include "driver.h"
 
+#if !defined (NDEBUG) && defined (LOG_LINK)
+#include "dbgif.h"
+#endif
+
 #if defined (LOG_MESSAGES)
 # define mlog		log
 #else
@@ -134,7 +138,7 @@ static unsigned			scheduler_suspend (unsigned long);
 static unsigned			scheduler_resume (void);
 static unsigned			controller_remove (void);
 static unsigned			controller_add (void);
-	
+
 struct capi_driver		capi_interface		= {
 
 	TARGET,
@@ -663,7 +667,12 @@ static int relocate_and_load (
 	unsigned	image_len,
 	unsigned long *	ptx, 
 	unsigned long *	prx
-) {
+#if defined (LOG_LINK)
+	,
+	unsigned long *	pstart,
+	unsigned long * ptrace
+#endif
+	) {
 	int		res;
 	int		size, align;
 	unsigned	usize;
@@ -703,6 +712,14 @@ static int relocate_and_load (
 	CHECK( lib_load_symtab (fw) );
 	CHECK( lib_findsym (fw, "_g_pToTMTransferInfoBase", ptx) );
 	CHECK( lib_findsym (fw, "_g_pToPCTransferInfoBase", prx) );
+#if defined (LOG_LINK)
+	if (IMAGE_OK != lib_findsym (fw, "_g_pDbgMemStart", pstart)) {
+		*pstart = 0;
+	}
+	if (IMAGE_OK != lib_findsym (fw, "_g_pDbgMemTrace", ptrace)) {
+		*ptrace = 0;
+	}
+#endif
 	res = lib_load (fw);
 	lib_remove (&fw);
 	return res;
@@ -739,6 +756,45 @@ static void start_ware (card_p cp) {
 	log (__FUNCTION__ "(%d) : 2: 0x%08x\n", __LINE__, val);
 } /* start_ware */
 
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+#if defined (TM_TIMESTAMP)
+
+#define TMTIMER_PER_SECOND      (1000*1000)
+#define TMTIMER_PER_SEC_2       (TMTIMER_PER_SECOND/2)
+#define DIVISOR                 ((IMAGE_FREQUENCY+TMTIMER_PER_SEC_2)/TMTIMER_PER_SECOND)
+#define	WRAP			(((unsigned) -1)/DIVISOR)
+
+#define	RX_TIMER		200	/* s */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+unsigned get_tm_timer (void) {
+	unsigned timer;
+	
+	assert (capi_card != NULL);
+	assert (DIVISOR > 0);
+	timer = PEEK (capi_card->io_base + TIMER2_TVALUE) / DIVISOR;
+	return timer;
+} /* get_tm_timer */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+static unsigned timer_status (unsigned start, unsigned now) {
+	unsigned	diff;
+
+	info (now >= start);
+	if (now >= start) {
+		diff = now - start;
+	} else {
+		info (WRAP >= start);
+		diff = WRAP - start + now;
+	}
+	return diff;
+} /* timer_status */
+
+#endif
+
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
 void * map_addr_from_tm_ptr (card_p cp, unsigned long X) {
@@ -773,6 +829,10 @@ int __entry load_ware (struct capi_ctr * ctrl, capiloaddata * ware) {
 	unsigned	count;
 #if !defined (NDEBUG)
 	unsigned	biu_ctl;
+#if defined (LOG_LINK)
+	unsigned long	ustart;
+	unsigned long	utrace;
+#endif
 #endif
 	assert (ware);
 	assert (ctrl);
@@ -802,7 +862,11 @@ int __entry load_ware (struct capi_ctr * ctrl, capiloaddata * ware) {
 	assert (PEEK (cp->io_base + MMIO_BASE) == cp->addr_mmio);
 	
 	utx = urx = 0;
+#if !defined (LOG_LINK)
 	if (IMAGE_OK != relocate_and_load (cp, buf, ware->firmware.len, &utx, &urx)) {
+#else
+	if (IMAGE_OK != relocate_and_load (cp, buf, ware->firmware.len, &utx, &urx, &ustart, &utrace)) {
+#endif
 		error ("Error: Firmware could not be relocated.\n");
 		return -EIO;
 	}
@@ -838,6 +902,10 @@ int __entry load_ware (struct capi_ctr * ctrl, capiloaddata * ware) {
 	assert (rx_info == NULL);
 	tx_info = (xfer_base_p) deref_map_addr_from_tm_ptr (cp, utx);
 	rx_info = (xfer_base_p) deref_map_addr_from_tm_ptr (cp, urx);
+#if !defined (NDEBUG) && defined (LOG_LINK)
+	dbg_setup (cp, ustart, utrace);
+	dbg_start ();
+#endif
 	load_flag = 1;
 	start_stack (cp);	
 	if (ctrl_context[0].ctrl != NULL) {
@@ -870,6 +938,9 @@ void __entry reset_ctrl (struct capi_ctr * ctrl) {
 		}
 	}
 	stop_stack (cp);
+#if !defined (NDEBUG) && defined (LOG_LINK)
+	dbg_stop ();
+#endif
 	reset_card (cp); 
 	load_flag = 0;
 	if (ctrl_context[0].ctrl != NULL) {
@@ -1733,6 +1804,9 @@ static void rx_dpc (void * data) {
 	void *			bufp;
 	xfer_info_t		info;
 	int			pending;
+#if defined (TM_TIMESTAMP)
+	unsigned		deltaT;
+#endif
 	
 	is_open = atomic_read (&link_open);
 	info (is_open);
@@ -1760,6 +1834,11 @@ static void rx_dpc (void * data) {
 	if (NULL != fp) {
 		cp->rx_func = NULL;
 		cp->rx_buffer = NULL;
+#if defined (TM_TIMESTAMP)
+		if ((deltaT = timer_status (info.stamp, get_tm_timer ())) > RX_TIMER) {
+			log ("RX timer violation [%us]: %u\n", RX_TIMER, deltaT);
+		}
+#endif
 		(*fp) (bufp, info.length, cp->rx_ctx);
 		kick_scheduler ();
 	}
@@ -1770,42 +1849,47 @@ static void rx_dpc (void * data) {
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
 static void irq_handler (int irq, void * args, struct pt_regs * regs) {
-	int		tx_flag;
-	int		rx_flag;
 	unsigned long	flags;
+	int		tx_flag	= 0;
+	int		rx_flag	= 0;
+#if !defined (NDEBUG)
+	int		count	= 0;
+#endif
 	
 	UNUSED_ARG (irq);
 	UNUSED_ARG (regs);
 	if (capi_card != (card_p) args) {
 		return;
 	}
-	flags = PEEK (capi_card->io_base + INT_CTL);
-	
-	if (0 == (flags & CARD_PCI_INT_ASSERT)) {
-		return;
-	}
-	assert (0 != (flags & CARD_PCI_INT_ENABLE));
-	assert (0 != (flags & CARD_PCI_INT_ISASSERTED));
-	
-	flags &= ~(CARD_PCI_INT_ASSERT | CARD_PCI_INT_ISASSERTED);
-	POKE (capi_card->io_base + INT_CTL, flags);
-	assert ((PEEK (capi_card->io_base + INT_CTL) 
+	while (0 != ((flags = PEEK (capi_card->io_base + INT_CTL)) & CARD_PCI_INT_ASSERT)) {
+#if !defined (NDEBUG)
+		++count;
+		assert (count < 3);
+#endif
+		assert (0 != (flags & CARD_PCI_INT_ENABLE));
+		assert (0 != (flags & CARD_PCI_INT_ISASSERTED));
+		flags &= ~(CARD_PCI_INT_ASSERT | CARD_PCI_INT_ISASSERTED);
+		POKE (capi_card->io_base + INT_CTL, flags);
+		assert ((PEEK (capi_card->io_base + INT_CTL) 
 				& ~(CARD_PCI_INT_ASSERT | CARD_PCI_INT_ISASSERTED)) != 0);
 
-	if (!atomic_read (&link_open)) {
-		return;
+		if (!atomic_read (&link_open)) {
+			return;
+		}
+
+		tx_flag = PEEK (capi_card->io_base + XFER_TOTM_STATUS) == TM_READY;
+		rx_flag = PEEK (capi_card->io_base + XFER_TOPC_STATUS) == TM_READY;
+		if (tx_flag) {
+			POKE (capi_card->io_base + XFER_TOTM_STATUS, PC_PENDING); 
+			queue_task (&tq_tx_dpc, &tq_private); 
+		}
+		if (rx_flag) {
+			POKE (capi_card->io_base + XFER_TOPC_STATUS, PC_PENDING);
+			queue_task (&tq_rx_dpc, &tq_private);
+		}
 	}
 
-	tx_flag = PEEK (capi_card->io_base + XFER_TOTM_STATUS) == TM_READY;
-	rx_flag = PEEK (capi_card->io_base + XFER_TOPC_STATUS) == TM_READY;
-	if (tx_flag) {
-		POKE (capi_card->io_base + XFER_TOTM_STATUS, PC_PENDING); 
-		queue_task (&tq_tx_dpc, &tq_private); 
-	}
-	if (rx_flag) {
-		POKE (capi_card->io_base + XFER_TOPC_STATUS, PC_PENDING);
-		queue_task (&tq_rx_dpc, &tq_private);
-	}
+	info (0 == (PEEK (capi_card->io_base + INT_CTL) & CARD_PCI_INT_ASSERT));
 	if (tx_flag || rx_flag) {
 		queue_task (&tq_dpc, &tq_immediate); 
 		mark_bh (IMMEDIATE_BH);	
@@ -1979,6 +2063,7 @@ int LinkTxMessage (void * msg, xfer_len_t len, completer_func_t cfunc, void * ct
 	capi_card->tx_ctx    = ctx;
 	memcpy (capi_card->tx_dmabuf, msg, len);
 	xfer_info_write (tx_info, 0, TM_PENDING, capi_card->tx_dmabuf_b, len);
+
 	return trigger_tx_interrupt ();
 } /* LinkTxMessage */
 
